@@ -150,7 +150,7 @@ couchdb_database_class_init (CouchdbDatabaseClass *klass)
 			      NULL, NULL,
 			      g_cclosure_marshal_VOID__OBJECT,
 			      G_TYPE_NONE, 1,
-			      G_TYPE_OBJECT);
+			      COUCHDB_TYPE_DOCUMENT);
 	couchdb_database_signals[DOCUMENT_UPDATED] =
 		g_signal_new ("document_updated",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -159,7 +159,7 @@ couchdb_database_class_init (CouchdbDatabaseClass *klass)
 			      NULL, NULL,
 			      g_cclosure_marshal_VOID__OBJECT,
 			      G_TYPE_NONE, 1,
-			      G_TYPE_OBJECT);
+			      COUCHDB_TYPE_DOCUMENT);
 	couchdb_database_signals[DOCUMENT_DELETED] =
 		g_signal_new ("document_deleted",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -180,13 +180,12 @@ couchdb_database_init (CouchdbDatabase *database)
 static CouchdbDocument *
 create_document_from_json_object (JsonObject *json_object)
 {
-	CouchdbDocument *document;
+	CouchdbDocument *document = NULL;
 
-	if (g_str_has_prefix (json_object_get_string_member (json_object, "id"),
-			      "_design/")) {
+	if (json_object_has_member (json_object, "_id") &&
+	    g_str_has_prefix (json_object_get_string_member (json_object, "_id"), "_design/")) {
 		document = COUCHDB_DOCUMENT (couchdb_design_document_new ());
-		couchdb_document_set_from_json_object (document,
-						       json_object_get_object_member (json_object, "doc"));
+		couchdb_document_set_from_json_object (document, json_object);
 	} else {
 		if (json_object_has_member (json_object, "record_type")) {
 			if (g_strcmp0 (json_object_get_string_member (json_object, "record_type"),
@@ -243,6 +242,8 @@ call_all_docs_uri (CouchdbSession *session, const char *url, JsonParser *parser,
 	return rows;
 }
 
+#ifdef COUCHDB_ENABLE_DEPRECATED
+
 /**
  * couchdb_database_list_documents:
  * @database: A #CouchdbDatabase object
@@ -288,6 +289,9 @@ couchdb_database_list_documents (CouchdbDatabase *database, GError **error)
 				json_object_get_string_member (
 					json_object_get_object_member (doc, "value"),
 					"rev"));
+			if (!doc_info)
+				continue;
+
 			doclist = g_slist_append (doclist, doc_info);
 		}
 	}
@@ -297,6 +301,7 @@ couchdb_database_list_documents (CouchdbDatabase *database, GError **error)
 
 	return doclist;
 }
+#endif 
 
 /**
  * couchdb_database_get_design_documents:
@@ -334,7 +339,7 @@ couchdb_database_get_design_documents (CouchdbDatabase *database, GError **error
 	if (rows != NULL) {
 		gint i;
 		for (i = 0; i < json_array_get_length (rows); i++) {
-			JsonObject *obj;
+			JsonObject *obj, *doc;
 			const gchar *id;
 			CouchdbDesignDocument *document;
 
@@ -342,13 +347,17 @@ couchdb_database_get_design_documents (CouchdbDatabase *database, GError **error
 			if (!obj)
 				continue;
 
-			id = json_object_get_string_member (obj, "_id");
+			doc = json_object_get_object_member (obj, "doc");
+			if (!doc)
+				continue;
+
+			id = json_object_get_string_member (doc, "_id");
 			if (!g_str_has_prefix (id, "_design/"))
 				continue;
 
 			document = couchdb_design_document_new ();
 			if (document != NULL) {
-				couchdb_document_set_from_json_object (COUCHDB_DOCUMENT (document), obj);
+				couchdb_document_set_from_json_object (COUCHDB_DOCUMENT (document), doc);
 				doclist = g_slist_append (doclist, document);
 			}
 		}
@@ -401,10 +410,12 @@ couchdb_database_get_all_documents (CouchdbDatabase *database, GError **error)
 			if (!obj)
 				continue;
 
-			document = create_document_from_json_object (json_object_get_object_member (obj, "doc"));
-			if (document != NULL) {
-				g_object_set (G_OBJECT (document), "database", database, NULL);
-				doclist = g_slist_append (doclist, document);
+			if (json_object_has_member (obj, "doc")) {
+				document = create_document_from_json_object (json_object_get_object_member (obj, "doc"));
+				if (document != NULL) {
+					g_object_set (G_OBJECT (document), "database", database, NULL);
+					doclist = g_slist_append (doclist, document);
+				}
 			}
 		}
 	}
@@ -444,6 +455,8 @@ couchdb_database_execute_view (CouchdbDatabase *database,
 	GSList *doclist = NULL;
 
 	g_return_val_if_fail (COUCHDB_IS_DATABASE (database), NULL);
+	g_return_val_if_fail (design_doc != NULL, NULL);
+	g_return_val_if_fail (view_name != NULL, NULL);
 
 	if (g_str_has_prefix (design_doc, "_design/")) {
 		url = g_strdup_printf ("%s/%s/%s/_view/%s",
@@ -506,10 +519,7 @@ couchdb_database_free_document_list (GSList *doclist)
 
 		doclist = g_slist_remove (doclist, data);
 
-		if (COUCHDB_IS_DOCUMENT (data))
-			g_object_unref (G_OBJECT (data));
-		else
-			couchdb_document_info_unref (data);
+		g_object_unref (G_OBJECT (data));
 	}
 }
 
@@ -634,10 +644,24 @@ couchdb_database_put_document (CouchdbDatabase *database,
 	body = couchdb_document_to_string (document);
 	parser = json_parser_new ();
 	if (id) {
-		char *encoded_docid;
+		char *encoded_docid, **special_name_parts;
+
+		/* Allow one sub-level in the ID string for i. e. design documents */
+		special_name_parts = *id == '_' ? g_strsplit (id, "/", 2) : NULL;
+		if (special_name_parts == NULL) {
+			encoded_docid = soup_uri_encode (id, NULL);
+			url = g_strdup_printf ("%s/%s/%s", couchdb_session_get_uri (database->priv->session), database->priv->dbname, encoded_docid);
+		}
+		else {
+			encoded_docid = soup_uri_encode (special_name_parts[1], NULL);
+			url = g_strdup_printf ("%s/%s/%s/%s",
+					       couchdb_session_get_uri (database->priv->session),
+					       database->priv->dbname,
+					       special_name_parts[0],
+					       encoded_docid);
+			g_strfreev (special_name_parts);
+		}
 
-		encoded_docid = soup_uri_encode (id, NULL);
-		url = g_strdup_printf ("%s/%s/%s", couchdb_session_get_uri (database->priv->session), database->priv->dbname, encoded_docid);		
 		send_ok = couchdb_session_send_message (database->priv->session, SOUP_METHOD_PUT, url, body, parser, error);
 
 		g_free (encoded_docid);
